# coding=utf-8
import logging
from justrpa.utils.config import load_merge_conf, load_override_conf
from justrpa.utils.time import get_report_period
from justrpa.errors import NoSiteError
from justrpa.task.executer import TaskExecuter
from justrpa.utils.file import save_json_file

class TaskHandler:
    def __init__(self):
        super().__init__()
        self.load_all_config()
    
    def load_all_config(self):
        self.logger = logging.getLogger("robot")
        self.env = self.load_env_variables()
        self.load_account()
        self.load_tasks()
        self.load_task_params()
        self.load_task_configs()
        self.init_account()
        self.generate_default_task_params()

    def load_env_variables(self):
        return load_merge_conf("env.json", "conf", "devdata")

    def load_conf(self, filename:str):
        return load_override_conf(filename, "conf", "devdata")

    def load_account(self):
        self.accounts = self.load_conf('account.json')
        if len(self.accounts)==0:
            self.accounts = [{"name":""}]

    def load_tasks(self):
        self.account_tasks = self.load_conf('task_sites.json')

    def load_task_params(self):
        self.task_params = self.load_conf('task_params.json')
    
    def load_task_configs(self):
        configfile = "task_configs.json" if "taskconfig" not in self.env else self.env["taskconfig"]
        self.task_configs = self.load_conf(configfile)
    
    def init_account(self):
        acc = self.env['account'] if 'account' in self.env else ""
        self.account = self.get_account(acc)
        self.env['account'] = self.account['name']
        self.env['account_info'] = self.account
    
    def get_task_params(self, account:str, taskname:str)->dict:
        # print(self.account_tasks)
        tasks = self.account_tasks[account]
        assert isinstance(tasks, dict) and (taskname in tasks), f"Task {taskname} is not found for account {account}"
        task_sites = tasks[taskname]
        account_sites = self.get_account(account)['sites']
        site_params = {}
        for site in task_sites:
            # print(f"site:{site}")
            if site not in account_sites:
                continue
            if isinstance(task_sites, dict):
                site_params[site] = task_sites[site]
            else:
                assert isinstance( self.task_params, dict) and (taskname in  self.task_params), f"Task {taskname} has no default params settings."
                site_params[site] = self.task_params[taskname]
        return site_params
    
    def generate_default_task_params(self):
        account_name = self.account['name']
        if account_name in self.account_tasks:
            return
        assert ("*" in self.account_tasks), f"task_sites.json error. No tasks defined for neither {account_name} or '*'."
        account_tasks = {}
        for taskname, sites_array in self.account_tasks["*"].items():
            site_for_account = [site for site in sites_array if site in self.account["sites"]]
            if len(site_for_account) > 0:
                account_tasks[taskname] = site_for_account
        self.account_tasks[account_name] = account_tasks
    
    def get_account(self, account_name:str="")->dict:
        if account_name=="":
            return self.accounts[0]
        else:
            selected = [acc for acc in self.accounts if acc['name']==account_name]
            assert len(selected)>0, f"Account {account_name} not found"
            return selected[0]

    def get_task_handler(self, taskname:str):
        assert taskname in self.task_configs, f"Task {taskname} config missing."
        assert "handler" in self.task_configs[taskname], f"Task {taskname} handler config missing."
        return self.task_configs[taskname]["handler"]

    def get_default_period_from_task_params(self, task_params):
        dr_task = ""
        for _, tasks in task_params.items():
            for params in tasks:
                if "date_range" in params:
                    dr_task = params["date_range"]
                    break
        dr = self.env['date_range'] if dr_task=="" and 'date_range' in self.env else dr_task
        dr = "LAST_MONTH" if dr=="" else dr
        execution_date = self.env['execution_date'] if 'execution_date' in self.env else ""
        return get_report_period(dr,execution_date=execution_date)

    def run(self, taskname:str=''):
        # use Environment variables.
        assert taskname!="" or "taskname" in self.env, f"taskname {taskname} is not presented either in cmd or env variables"
        taskname = self.env['taskname'] if taskname=='' else taskname
        task_params = self.get_task_params(self.account['name'] , taskname)
        report_period = self.get_default_period_from_task_params(task_params)
        self.env['report_period'] = report_period
        self.env['taskname'] = taskname
        try:
            self.logger.info(f"Before run task: {taskname}. params: {task_params}")
            if len(task_params) ==0:
                raise NoSiteError("No site specified for this task, please check settings.")
            self.do_task(taskname, task_params)
            self.logger.info(f"Run task completed.")
        except Exception as ex:
            self.logger.error(f"Run task failed.")
            self.logger.error(ex, exc_info=True, stack_info=True)
            self.save_task_erors(self.account['name'], report_period, taskname, ex)

    def get_env_variables(self, taskname:str):
        current = self.env.copy()
        if self.task_configs and taskname in self.task_configs:
            current.update(self.task_configs[taskname])
        return current

    def do_task(self, taskname, task_params):
        # print(f"taskname:{taskname} params:{task_params}")
        browser_name, method_name = self.get_task_handler(taskname)
        env = self.get_env_variables(taskname)
        t = TaskExecuter(taskname, browser_name, method_name, env, task_params)
        task_context = t.run()
        return task_context

    def get_error_msg(self, ex):
        msg_map = {
            "WindowNotFound": "机器人浏览器窗口被意外关闭,请重试.",
            "NoOpenBrowser": "机器人找不到可用浏览器窗口,请重试.",
            "BrowserNotFoundError": "RPA浏览器未打开,机器人无法启动,请打开后重试.",
            "NoSuchWindowException": "机器人浏览器窗口被意外关闭,请重试.",
            "BrowserNotAvailable": "RPA浏览器未打开,机器人无法启动,请打开后重试.",
            "NotLoggedInError": "账号未在浏览器登录,请登录后重试.",
            "LoginFailError": "自动登录失败,请检查账号密码设置,或者手动登录后重试.",
            "MFAFailError": "MFA(两步认证)失败,请检查相关设置,或者手动登录后重试.",
            "NoSiteError": "本任务没有关联站点,请检查相关设置"
        }
        ex_name = ex.__class__.__name__
        if ex_name not in msg_map:
            return str(ex)
        return msg_map[ex_name]

    def get_user_actions(self, ex):
        action_map = {
            "WindowNotFound": "Browser window has been closed, please open and retry task.",
            "NoOpenBrowser": "RPA Browser is not opened, please open and retry task.",
            "BrowserNotFoundError": "RPA Browser is not opened, please open and retry task.",
            "NoSuchWindowException": "Browser window has been closed, please open and retry task.",
            "BrowserNotAvailable": "RPA Browser is not opened, please open and retry task.",
            "NotLoggedInError": "Please login in the browser.",
            "LoginFailError": "Login fail, please check your login settings or login in the browser manually.",
            "MFAFailError": "Two factor authentication failed, please check your settings or do it in the browser manually",
            "NoSiteError": "No site has been set to this task, please check task settings."
        }
        ex_name = ex.__class__.__name__
        if ex_name not in action_map:
            return "Check and retry or contact support."
        return action_map[ex_name]

    def save_task_erors(self, account:str, report_period:str, taskname:str, err:Exception, output_dir:str='output/', filename:str='task_result.json'):
        result = {
            'account': account,
            'taskname': taskname,
            'report_period': report_period,
            'status': "Error",
            'error': self.get_error_msg(err),
            'action': self.get_user_actions(err)
        }
        save_json_file(result, output_dir, filename)


# coding=utf-8

from copy import deepcopy
from justrpa.utils.file import save_json_file

def get_klass(full_class_name:str):
    symbols = full_class_name.split(".")
    classname = symbols[-1]
    if len(symbols)==1:
        klass = globals()[classname]
    else:
        classmodule = ".".join(symbols[0:-1])
        mod = __import__(classmodule, fromlist=[classname])
        klass = getattr(mod, classname)
    return klass

def calc_metrics(total:dict, success:dict, failure:dict):
    metrics = {}
    total_suc = 0
    total_fail = 0
    for site, params in total.items():
        suc = len(success[site]) if site in success else 0
        fail = len(failure[site]) if site in failure else 0
        metrics[site] = {
            'success': suc,
            'fail': fail,
            'total': suc+fail
        }
        total_suc = total_suc + suc
        total_fail = total_fail + fail
    metrics['total'] = total_suc + total_fail
    metrics['success'] = total_suc
    metrics['fail'] = total_fail
    return metrics

def gen_task_result(account:str, report_period:str, taskname:str, total:dict, success:dict, failure:dict)->dict:
    metrics = calc_metrics(total, success, failure)
    status="Complete" if metrics['fail']==0 else "Partial"
    result = {
        'account': account,
        'taskname': taskname,
        'report_period': report_period,
        "period_type": "Monthly",
        'status': status,
        'metrics': metrics,
        'data':{
            'total':total,
            'success':success,
            'failure': failure
        }
    }
    return result

def save_tasks_summary(account:str,report_period:str,taskname:str, total:dict, success:dict, failure:dict, output_dir:str='output/', filename:str='task_result.json'):
    result = gen_task_result(account, report_period,taskname, total, success, failure)
    save_json_file(result, output_dir, filename)
    return result['metrics']

def save_fail_tasks(account:str, taskname:str, failure:dict, output_dir:str='output/', filename:str='task_failed.json'):
    result = {
        account:{
            taskname:failure
        }
    }
    save_json_file(result, output_dir, filename)

class TaskExecuter:
    def __init__(self, taskname:str, taskclass:str, taskmethod:str, env_variables:dict, task_params:dict):
        self.taskname = taskname
        self.taskclass = taskclass
        self.taskmethod = taskmethod
        self.env_variables = env_variables.copy()
        self.task_params = task_params.copy()
        self.init_task_class()

    def init_task_class(self):
        klass = get_klass(self.taskclass)
        self.taskinstance = klass()
        self.taskinstance.init(self.env_variables)
    
    def call_taskclass_method(self, method_name:str, task_context:dict):
        method = getattr(self.taskinstance, method_name)
        if task_context:
            return method(task_context)
        else:
            return method()
    
    def run_process_chain(self, task_context:dict, method_chain:list):
        for meth in method_chain:
            if callable(meth):
                task_context = meth(task_context)
            else:
                task_context = self.call_taskclass_method(meth, task_context)
        return task_context

    def run_tasks_by_site(self, task_context:dict)->dict:
        res_success = {}
        res_failure = {}
        task_params = task_context["all_task_params"]
        level_two_chain = ["before_task_run", "run_task", "after_task_run"]
        for site, site_reports in task_params.items():
            task_context["site"] = site
            task_context = self.call_taskclass_method("switch_site", task_context)
            success, failure = [],[]
            for params in site_reports:
                task_context["task_params"] = params
                task_context = self.run_process_chain(task_context, level_two_chain)
                if task_context["task_status"] == "success":
                    success.append(task_context["task_result"])
                else:
                    failure.append(params)
            res_success[site] = success
            res_failure[site] = failure
        task_context["res_success"] = res_success
        task_context["res_failure"] = res_failure
        return task_context
    
    def generate_task_outputs(self, task_context:dict)->dict:
        account = task_context["account"]
        total = task_context["all_task_params"]
        res_success = task_context["res_success"]
        res_failure = task_context["res_failure"]
        report_period = task_context["report_period"]
        taskname = task_context["taskname"]
        metrics = save_tasks_summary(account, report_period, self.taskname, total, res_success, res_failure)
        save_fail_tasks(account, taskname, res_failure)
        return task_context

    def run(self):
        task_context = {}
        task_context['all_task_params'] = self.task_params
        task_context['taskmethod'] =  self.taskmethod
        task_context['taskclass'] = self.taskclass
        task_context['taskname'] = self.taskname
        first_level_chain = ["before_run", "process_task_params", self.run_tasks_by_site, "after_run", self.generate_task_outputs]
        try:
            task_context = self.run_process_chain(task_context, first_level_chain)
        finally:
            task_context = self.call_taskclass_method("task_end", task_context)
        return task_context


import os
import json
import pandas as pd
import xlsxwriter

def check_file_exists(target_filename:str, ignore_ext:bool=False)->bool:
    target_filename = os.path.abspath(target_filename)
    if ignore_ext==True:
        dir_name = os.path.dirname(target_filename)
        filename = os.path.splitext(target_filename)[0]
        for file in os.listdir(dir_name):
            current_filename = os.path.splitext(file)[0]
            if current_filename == filename:
                return True
        return False
    else:
        return os.path.isfile(target_filename)

def create_dir_for_file(filename):
    dir = os.path.dirname(filename)
    if os.path.isdir(dir)==False:
        os.makedirs(dir, exist_ok=True)

def detect_encoding(file_name:str)->str:
    with open(file_name, 'rb') as f:
        rawdata = f.read()
    encodings = ['utf-8','CP932', 'shift-jis', 'euc-jp', 'iso-2022-jp', 'iso-8859-1']
    for encoding in encodings:
        try:
            rawdata.decode(encoding)
            return encoding
        except:
            pass
    return 'utf-8'

def merge_csv_files(files:list, output_file:str, header_row=1, delete=True):
    print(f"Merging {len(files)} files to {output_file}")
    encoding = detect_encoding(files[0])
    with open(output_file, 'w', encoding=encoding) as outfile:
        for idx, fname in enumerate(files):
            with open(fname, 'r', encoding=encoding) as infile:
                for idy, line in enumerate(infile):
                    if idx<1 or idy >=header_row:
                        outfile.write(line)
    print(f"Merged {len(files)} files to {output_file}")
    if delete:
        for fname in files:
            os.remove(fname)
    return output_file

def load_json_file(file_dir:str, filename:str)->dict:
    dir_path = dir_path = os.getcwd()
    dir_path = os.path.join(dir_path, file_dir)
    file_path = os.path.join(dir_path, filename)
    with open(file_path, 'r') as infile:
        return json.load(infile)

def save_json_file(result:dict, output_dir, filename:str):
    # dir_path = os.path.dirname(os.path.realpath(__file__))
    dir_path = os.getcwd()
    dir_path = os.path.join(dir_path, output_dir)
    if not os.path.exists(dir_path):
        os.makedirs(dir_path)
    file_path = os.path.join(dir_path, filename)
    with open(file_path, 'w') as outfile:
        json.dump(result, outfile, indent=4)

def load_csv(file_name:str)->pd.DataFrame:
    encoding = detect_encoding(file_name)
    try:
        df = pd.read_csv(file_name, encoding=encoding)
    except pd.errors.EmptyDataError:
        print(f"{file_name} is empty csv.")
        df = pd.DataFrame()
    return df

def save_dfs_tabs(df_list, sheet_list, file_name):
    writer = pd.ExcelWriter(file_name,engine='xlsxwriter')   
    for dataframe, sheet in zip(df_list, sheet_list):
        dataframe.to_excel(writer, sheet_name=sheet, startrow=0 , startcol=0, index=False)   
    writer.save()

def merge_csv_to_excel(csv_files, output_file, sheet_names=[]):
    dfs = [ load_csv(f) for f in csv_files if f != None]
    if len(sheet_names) == 0:
        sheet_names = [f"sheet{i}" for i in range(len(dfs))]
    save_dfs_tabs(dfs,sheet_names, output_file)
    return output_file


import os
import subprocess
import platform
import psutil

# todo: doublecheck: multiple platform handle 

def RunAsAdmin(path_to_file,*args):
    cmd = r'Powershell -Command "Start-Process "'+path_to_file+'"' + ' -ArgumentList @('+str(args)[1:-1]+')'+ ' -Verb RunAs"'
    print(cmd)
    os.system(cmd)

def RunByPowershell(path_to_file,*args):
    cmd = r'Powershell -Command "Start-Process "'+path_to_file+'"'+ ' -ArgumentList @('+str(args)[1:-1]+')"' 
    print(cmd)
    os.system(cmd)

def kill_process_as_admin(process_name:str):
    RunAsAdmin('taskkill','/F','/T','/IM', process_name)

def kill_process(process_name:str):
    cmd_map = {
        "Windows": f"taskkill /F /T /IM {process_name}",
        "Linux": f"killall -9 {process_name}",
        "Darwin": f"pkill -f  {process_name}"
    }
    system = platform.system()
    cmd = cmd_map[system]
    print(f"system:{system}, cmd:{cmd}")
    subprocess.call(cmd, shell=True)

def kill_all_chrome():
    proc_map = {
        "Windows": ["CHROME.EXE", "CHROMEDRIVER.EXE"],
        "Linux": ["chrome", "chromedriver"],
        "Darwin": ["Chrome", "chromedriver"]
    }
    system = platform.system()
    for proc in proc_map[system]:
        kill_process(proc)
        if check_process_exists(proc):
            print(f"warning: process {proc} still exists. kill failed.")
            if system == "Windows":
                print("Retry with windows runas admin")
                kill_process_as_admin(proc)
                if check_process_exists(proc):
                    print(f"warning: process {proc} still exists. kill failed.")
                else:
                    print(f"kill {proc} success")
        else:
            print(f"kill {proc} success")

def does_chrome_exists():
    proc_map = {
        "Windows": ["CHROME.EXE"],
        "Linux": ["chrome"],
        "Darwin": ["Chrome"]
    }
    system = platform.system()
    any_exists = False
    for proc in proc_map[system]:
        any_exists = any_exists or check_process_exists(proc)
    return any_exists

def clean_up_chrome():
    kill_all_chrome()
    return does_chrome_exists()==False

def check_process_exists(exename:str):
    for proc in psutil.process_iter(['pid', 'name']):
        # This will check if there exists any process running with executable name
        if proc.info['name'].lower() == exename.lower():
            return True
    return False

def start_chrome_windows(clean=False):
    if clean:
        cleaned = clean_up_chrome()
        if cleaned==False:
            print(f"!!warning: cleanup failed ...")
            print(f"!!warning: can't start new chrome due to cleanup failed ...")
            return False
    if does_chrome_exists():
        print(f"Chrome already exists. Skip launch new chrome.")
        return True
    # cmd =  "start Chrome --remote-debugging-port=9222 --kiosk-printing --start-maximized --disable-dev-shm-usage --disable-web-security --allow-running-insecure-content"
    # print(f"run cmd:{cmd}")
    # subprocess.call(cmd, shell=True)
    RunByPowershell('Chrome', '--remote-debugging-port=9222', '--kiosk-printing', '--start-maximized', '--disable-dev-shm-usage', '--disable-web-security', '--allow-running-insecure-content')
    return does_chrome_exists()==True




import os
import time
from typing import List, Optional, Union
from io import BytesIO
from PIL import Image
# RPA library imports
from RPA.core import webdriver
from RPA.Browser.Selenium import Selenium
from selenium.webdriver import FirefoxProfile
from selenium.webdriver.remote.webelement import WebElement
# Local imports
from justrpa.browser.chrome import start_chrome_windows
from justrpa.browser.firefox import get_firefox_profile_path, FILE_CONTENT_TYPE
from justrpa.utils.file import create_dir_for_file
from justrpa.errors import BrowserNotAvailable

# RPA Browser Selenium Extension; adds common browser methods.
class SeleniumExt(Selenium):  
    def sleep_short(self):
        self.sleep(2)
    
    def sleep_medium(self):
        self.sleep(4)
    
    def sleep_long(self):
        self.sleep(8)
    
    def sleep_very_long(self):
        self.sleep(20)

    def sleep(self, seconds:int):
        time.sleep(seconds)

    def get_driver_path(self, browser:str)->str:
        def is_tool(name):
            """Check whether `name` is on PATH and marked as executable."""
            from shutil import which
            return which(name) is not None
        path_cache = webdriver.cache(browser)
        if path_cache:
            return path_cache
        # Try to download webdriver
        try:
            path_download = webdriver.download(browser)
            return path_download
        except Exception as e:
            self.logger.error(f"Failed to download webdriver: {e}")
            pass
        return None

    def get_firefox_profile(self, download_dir:str=None):
        ff_profile_path = get_firefox_profile_path()
        self.logger.info(f"Firefox profile: {ff_profile_path} ")
        profile = FirefoxProfile(ff_profile_path)
        profile.set_preference("browser.download.folderList", 2)
        profile.set_preference("browser.link.open_newwindow", 1)
        profile.set_preference("browser.download.manager.showWhenStarting",False)
        if download_dir:
            dl_dir = os.path.abspath(download_dir)
            self.logger.info(f"download_dir:{dl_dir}")
            profile.set_preference("browser.download.dir", dl_dir)
        profile.set_preference("browser.helperApps.alwaysAsk.force", False)
        profile.set_preference("browser.helperApps.neverAsk.saveToDisk", FILE_CONTENT_TYPE)
        profile.set_preference("browser.download.manager.closeWhenDone", True)
        profile.set_preference("pdfjs.disabled", True)
        return profile

    def open_firefox_browser(self, url:str, download_dir:str=None)->None:
        driver_path = self.get_driver_path("firefox")
        self.logger.info(f"driver path: {driver_path}")
        ff_profile_path = self.get_firefox_profile(download_dir)
        self.logger.info(f"Firefox profile: {ff_profile_path.profile_dir} ")
        self.logger.info(f"Firefox profile: {ff_profile_path.path} ")
        # self.open_browser(url,browser="firefox", ff_profile_dir="/Users/yanjing/Library/Application Support/Firefox/Profiles/kpzh8uvf.default-release")
        self.open_browser(url,browser="firefox", ff_profile_dir=ff_profile_path, executable_path=f"{driver_path}", service_log_path='output/gecko.log')

    def open_new_tab(self, url):
        self.logger.info(f"open new tab: {url}")
        num_tabs = len(self.driver.window_handles)
        self.driver.execute_script("window.open('about:blank', '_blank');")
        num_tabs_after = len(self.driver.window_handles)
        self.logger.info(f"num_tabs: {num_tabs}, num_tabs_after: {num_tabs_after}")
        self.new_tab= num_tabs_after>num_tabs 
        self.driver.switch_to.window(self.driver.window_handles[-1])
        self.driver.get(url)
    
    def close_tab(self):
        opened = self.new_tab if hasattr(self, 'new_tab') else False
        self.logger.info(f"opened tab: {opened}. ")
        if opened:
            self.logger.info(f"close tab: {self.driver.current_url}")
            self.driver.close()

    def prepare_browser(self, url:str)->None:
        self.logger.info(f"browser:{self.browser}, New browser: {self.new_browser}, {self.browser_info}")
        handler = {
            0:self.attach_default_browser,
            1:self.open_new_browser,
            2:self.open_after_attach_fail
        }
        # info(f"{handler[self.new_browser]}")
        handler[self.new_browser](url)
        #self.attach_default_browser(url)

    def open_new_browser(self, url:str)->None:
        browser_info = self.browser_info
        browser_name = browser_info['name']
        if browser_name == 'chrome':
            # profile_path = "/Users/yanjing/Library/Application Support/Google/Chrome/"
            #options.AddArgument("--ignore-certificate-errors");
            #options.AddArgument("--disable-popup-blocking");
            # info(f"download_dir:{download_dir}")
            # prefs = {
            #     "download.default_directory": download_dir,
            #     "download.prompt_for_download": False,
            #     "download.directory_upgrade": True,
            # }
            self.logger.info(f"open chrome with default profile")
            if hasattr(self, "download_dir"):
                self.download_dir = os.path.abspath(self.download_dir)
                self.set_download_directory(self.download_dir)
            self.open_chrome_browser(url, use_profile=False, maximized=True, headless=self.headless)
            self.logger.info(f"Opended browser")
            self.go_to(url)
            self.logger.info(f"Go to {url}")
            # self.open_chrome_browser(url, use_profile=browser_info["use_profile"],profile_path=browser_info["profile_path"])
        elif browser_name == 'default':
            self.open_available_browser(url)
        elif browser_name == 'firefox':
            self.open_firefox_browser(url, self.download_dir)
            self.set_selenium_speed(0.2)
        else:
            self.open_available_browser(url, browser_selection=browser_name)   
        self.logger.info(f"Opened new browser: {browser_name}. ")

    def open_after_attach_fail(self, url:str)->None:
        try:
            self.check_browser_and_attach(url, clean=False)
        except BrowserNotAvailable as ex:
            self.logger.error(ex, stack_info=True)  
            self.check_browser_and_attach(url, clean=True)
    
    def check_browser_and_attach(self, url:str, clean=False):
        result = start_chrome_windows(clean)
        if result:
            self.sleep_short()
            self.attach_default_browser(url)
        else:
            self.logger.error("Can't connect to Chrome. And start failed.")

    def attach_default_browser(self,url:str):
        try:
            self.attach_chrome_browser(self.attach_port)
            self.set_selenium_timeout(20)
        except Exception as ex:
            self.logger.error(ex, stack_info=True)  
            raise BrowserNotAvailable("Attach browser failed. Browser is not opened or not available for connect.") 
        self.logger.info(f"Attached to browser.")
        self.open_new_tab(url)

    def cleanup(self):
        self.sleep_short()
        self.close_tab()
        if self.new_browser == 1:
            self.logger.info(f"Opened new browser, will close it.")
            self.close_browser()

    ################################################################
    ## Screenshot functions
    def screenshot_fullpage(self, file, scroll_delay=0.3):
        driver = self.driver
        device_pixel_ratio = driver.execute_script('return window.devicePixelRatio')
        total_height = driver.execute_script('return document.body.parentNode.scrollHeight')
        viewport_height = driver.execute_script('return window.innerHeight')
        total_width = driver.execute_script('return document.body.offsetWidth')
        viewport_width = driver.execute_script("return document.body.clientWidth")
        assert(viewport_width == total_width)
        # scroll the page, take screenshots and save screenshots to slices
        offset = 0
        slices = {}
        while offset < total_height:
            if offset + viewport_height > total_height:
                offset = total_height - viewport_height
            driver.execute_script('window.scrollTo({0}, {1})'.format(0, offset))
            time.sleep(scroll_delay)
            img = Image.open(BytesIO(driver.get_screenshot_as_png()))
            slices[offset] = img
            offset = offset + viewport_height
        # combine image slices
        # info(f"{device_pixel_ratio} {total_height} {viewport_height} {total_width} {viewport_width}")
        stitched_image = Image.new('RGB', (int(total_width * device_pixel_ratio), int(total_height * device_pixel_ratio)))
        for offset, image in slices.items():
            stitched_image.paste(image, (0, int(offset * device_pixel_ratio)))
        stitched_image.save(file)

    def screenshot_to_dest(self, filepath:str, full_page:bool=False):
        create_dir_for_file(filepath)
        if full_page:
            self.screenshot_fullpage(file=filepath)
        else:
            self.screenshot(filename=filepath)

    ################################################################
    # JS functions
    def scroll_to_element(self, locator: Union[WebElement, str]):
        element = self.find_element(locator)
        self.driver.execute_script("arguments[0].scrollIntoView();", element)
    
    def zoom_in(self, ratio=1):
        js = f"document.body.style.zoom = '{ratio}'"
        self.driver.execute_script(js)
    
    def input_element_text_by_js(self, locator: Union[WebElement, str], text:str):
        element = self.find_element(locator)
        script = f'arguments[0].value="{text}";arguments[0].dispatchEvent(new Event("change"));'
        self.driver.execute_script(script, element)

    def click_element_by_js(self, locator: Union[WebElement, str]):
        self.wait_until_element_is_visible(locator,20)
        self.click_element_by_js_directly(locator)

    def click_element_by_js_directly(self, locator: Union[WebElement, str]):
        element = self.find_element(locator)
        script = f'arguments[0].click();'
        self.driver.execute_script(script, element)

    def hide_element_by_js(self, locator: Union[WebElement, str]):
        if self.does_page_contain_element(locator)==False:
            return
        element = self.find_element(locator)
        script = f"arguments[0].innerHTML = '';"
        self.driver.execute_script(script, element)

    ################################################################
    # Simple combine functions    
    def focus_and_input_text(self,locator: Union[WebElement, str], text:str):
        self.mouse_over(locator)
        self.clear_element_text(locator)
        #self.input_text(locator, text)
        self.input_element_text_by_js(locator, text)
    
    def focus_and_click_element(self, locator: Union[WebElement, str]):
        self.mouse_over(locator)
        self.click_element(locator)
    
    def focus_and_select_checkbox(self, locator: Union[WebElement, str]):
        self.mouse_over(locator)
        self.select_checkbox(locator)

    ##################################################################
    # actions
    def action_screenshot(self, filepath:str, full_page:bool=False, hide_locators:List[str]=[]):
        for locator in hide_locators:
            self.hide_element_by_js(locator)
        self.screenshot_to_dest(filepath, full_page)


def get_site_short(site:str)->str:
    site_map = {
        "United States":"US",
        "United Kingdom":"UK",
        "Japan":"JP",
        "Germany":"DE",
        "Italy":"IT",
        "Canada":"CA",
        "Mexico":"MX",
        "Australia":"AU",
        "France":"FR",
        "Spain":"ES",
        "Netherlands":"NL",
        "Sweden":"SE",
        "Switzerland":"CH",
        "Poland":"PL"
    }
    if site not in site_map:
        return site.replace(" ", "")
    return site_map[site]

def get_amazon_home_url(site:str)->str:
    url_map = {
        "United States":"https://sellercentral.amazon.com/",
        "United Kingdom":"https://sellercentral.amazon.co.uk/",
        "Japan":"https://sellercentral-japan.amazon.com/"
    }
    if site in ["Italy", "United Kingdom", "France", "Spain", "Poland", "Sweden","Netherlands"]:
        site = "United Kingdom"
    elif site in ["Germany", "Canada", "Mexico", "United States"]:
        site = "United States"
    if site not in url_map:
        site = "United States"
    return url_map[site]

def get_site_short_cn(site:str)->str:
    site_map = {
        "United States":"美国",
        "Japan":"日本",
        "United Kingdom":"英国",
        "Canada":"加拿大",
        "Germany":"德国",
        "France":"法国",
        "Italy":"意大利"
    }
    if site not in site_map:
        return get_site_short(site)
    return site_map[site]


import logging
import logging.config
import sys
import os
import json
from .errors import *

__version__ = "0.1.3"
DEFAULT_LOGGER_NAME = "robot"

def create_logger(log_level:int, formatter:str, output_stream, name=None) -> logging.Logger:
    log_level = log_level # to set log level
    formatter = logging.Formatter(formatter)
    my_logger = logging.getLogger(name)
    my_logger.setLevel(log_level)
    log_handler = logging.StreamHandler(output_stream)
    log_handler.setLevel(log_level)
    log_handler.setFormatter(formatter)
    my_logger.addHandler(log_handler)
    return my_logger

def create_default_logger()->logging.Logger:
    Default_Log_level = 10
    Default_Formatter = "[%(asctime)s] [%(levelname)s] [%(message)s]"
    Default_Output = sys.stdout
    return create_logger(Default_Log_level, Default_Formatter, Default_Output, DEFAULT_LOGGER_NAME)

def load_logging_config():
    dir_path = os.getcwd()
    file_path = os.path.join(dir_path, "conf/logging.json")
    if os.path.isfile(file_path) == False:
        return {}
    with open(file_path, encoding="utf-8") as fp:
            conf = json.load(fp)
    return conf

def _init_logger():
    log_config = load_logging_config()
    if log_config:
        logging.config.dictConfig(log_config)
    if DEFAULT_LOGGER_NAME not in logging.Logger.manager.loggerDict.keys():
        create_default_logger()

_init_logger()


import os
import datetime
from justrpa.utils.site import get_site_short, get_site_short_cn
#
# filepath_template
# module_filepath_template
# replace_template
class ReportPath:
    DEFAULT_TEMPLATE = "{account}/{site}/{taskname}/{account}-{site}-{filetag}-{timetag}.csv"
    def try_get_attr(self, var_name, default_value):
        return getattr(self, var_name) if hasattr(self, var_name) else default_value

    def get_template(self, site:str, module:str)->str:
        replaced = self.get_replace_template(site, module)
        if replaced:
            return replaced
        module_template = self.get_module_filepath_template(module)
        if module_template:
            return module_template
        return self.get_filepath_template()
    
    def get_filepath_template(self)->str:
        return self.try_get_attr('filepath_template', self.DEFAULT_TEMPLATE)

    def get_module_filepath_template(self, module:str)->str:
        if module=="":
            return None
        templates = self.try_get_attr('module_filepath_template', {})
        if  module not in templates:
            return None
        return templates[module]

    def get_replace_template(self, site, module):
        if site=="" or module == "":
            return None
        replace_templates = self.try_get_attr('replace_template', {})
        if site not in replace_templates:
            return None
        site_templates = replace_templates[site]
        self.logger.info(f"replaced site templates found for {site} - {module}")
        if 'module_filepath_template' in site_templates and module in site_templates['module_filepath_template']:
            return site_templates['module_filepath_template'][module]
        elif 'filepath_template' in site_templates:
            return site_templates['filepath_template']
        else:
            return None

    def build_filename(self, template:str="", **kwargs)->str:
        """
        Support kwargs:
        - account
        - site
        - taskname
        - filetag
        - timetag
        """
        params = kwargs.copy()
        if template=="":
            site = params["site"] if "site" in params else ""
            module = params["module"] if "module" in params else ""
            template = self.get_template(site, module)
        if "site" in params:
            params["site_short"] = get_site_short(params["site"])
            params["site_short_cn"] = get_site_short_cn(params["site"])
            params["site"] = params["site"].replace(" ", "")
        if "account" not in params:
            params['account'] = self.account
        if params['account'] != None:
            params['account_trim'] = params['account'].split("-")[0]
        if "taskname" not in params:
            params['taskname'] = self.taskname
        if "today" not in params:
            params['today'] = datetime.datetime.now().strftime("%Y-%m-%d")
        if "timetag" not in params:
            params['timetag'] = params['start_date'][:7] if 'start_date' in params else ''
        if "filetag" not in params:
            params['filetag'] = ""
        filepath = template.format(**params)
        if "fileext" in params:
            filepath = filepath.rsplit('.', 1)[0] + params['fileext']
        return filepath
    
    def build_filepath(self, template:str="", **kwargs)->str:
        dir_path = os.getcwd()
        dir_path = os.path.join(dir_path, self.target_dir)
        filename = self.build_filename(template, **kwargs)
        return os.path.join(dir_path, filename)



class JustRpaException(Exception):
    pass

class BrowserNotAvailable(JustRpaException):
    pass

class NotLoggedInError(JustRpaException):
    pass

class LoginFailError(JustRpaException):
    pass

class MFAFailError(JustRpaException):
    pass

class GotoPageFailError(JustRpaException):
    pass

class ReportOptionError(JustRpaException):
    pass

class NoDownloadLinkError(JustRpaException):
    pass

class FileDownloadError(JustRpaException):
    pass

class NoReportError(JustRpaException):
    pass

class ProgramError(JustRpaException):
    pass

class NoSiteError(JustRpaException):
    pass

class ParameterError(JustRpaException):
    pass

class TooManyExceptionsError(JustRpaException):
    pass

class ErrorCounter:
    def __init__(self, limit=5):
        self.limit = limit
        self.counter_map = {}

    def reset(self, category=""):
        self.counter_map[category] = 0
    
    def add(self, category=""):
        if category not in self.counter_map:
            self.counter_map[category] = 0
        self.counter_map[category] = self.counter_map[category] + 1
        self.check(category)
    
    def get(self, category=""):
        return self.counter_map[category] if category in self.counter_map else 0
    
    def check(self, category=""):
        if self.get(category) >= self.limit:
            raise TooManyExceptionsError(f"Too many exceptions. Category:{category}, Limit:{self.limit}")



# Always prefer setuptools over distutils
import re
from setuptools import setup, find_packages

# To use a consistent encoding
from codecs import open
from os import path

# The directory containing this file
HERE = path.abspath(path.dirname(__file__))

with open(path.join(HERE, 'justrpa', '__init__.py')) as f:
    VERSION = re.search('\n__version__ = "(.*)"', f.read()).group(1)

# Get the long description from the README file
with open(path.join(HERE, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

with open(path.join(HERE, 'requirements.txt')) as f:
    REQUIREMENTS = f.read().splitlines()

# This call to setup() does all the work
setup(
    name="justrpa",
    version=VERSION,
    description="Just RPA Common Library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://justrpa.com/",
    author="loop",
    author_email="jing.yan.cn@gmail.com",
    license="MIT",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: OS Independent"
    ],
    packages=find_packages(include=["justrpa", "justrpa.*"]),
    include_package_data=True,
    install_requires=REQUIREMENTS
)



import os
import json
import requests

def merge_conf(origin:dict, custom:dict)->dict:
    for key, value in custom.items():
        origin[key] = value
    return origin

def load_conf(filename):
    # dir_path = os.path.dirname(os.path.realpath(__file__))
    dir_path = os.getcwd()
    file_path = os.path.join(dir_path, filename)
    if os.path.isfile(file_path) == False:
        return {}
    with open(file_path, encoding="utf-8") as fp:
            conf = json.load(fp)
    return conf

def load_override_conf(filename, default_dir="conf", override_dir="devdata"):
    path_origin, path_override = [ os.path.join(d, filename) for d in [default_dir, override_dir]]
    conf_default, conf_override = [ load_conf(p) for p in [path_origin, path_override] ]
    return conf_override if len(conf_override) > 0 else conf_default

def load_merge_conf(filename, default_dir="conf",override_dir="devdata"):
    path_origin, path_override = [ os.path.join(d, filename) for d in [default_dir, override_dir]]
    conf_default, conf_override = [ load_conf(p) for p in [path_origin, path_override] ]
    return merge_conf(conf_default, conf_override)

def load_json_from_url(url:str):
    res = requests.get(url)
    data = json.loads(res.text)
    return data


import os
import time
import datetime
import shutil
import hashlib
import logging
from .selenium import SeleniumExt
from .report import ReportPath
from justrpa.errors import FileDownloadError
from justrpa.utils.file import save_json_file,create_dir_for_file
from justrpa.utils.time import get_date_range_by_param

class TaskBrowser(SeleniumExt, ReportPath):
    def _init_variables(self):
        self.selenium_speed = 1
        self.account = ""
        self.home_url = "about:blank"
        self.filepath_template = ""
        self.report_period = ""
        self.execution_date = ""
        self.task_output_dir = "output/"

    def init(self, env:dict):
        self.logger = logging.getLogger("robot")
        self._init_variables()
        self.set_env_variables(env)
        self.set_account_info(env)
        self.set_browser_settings(env)
        self.set_selenium_speed(self.selenium_speed)

    def set_env_variables(self,*initial_data, **kwargs):
        for dictionary in initial_data:
            for key in dictionary:
                setattr(self, key, dictionary[key])
        for key in kwargs:
            setattr(self, key, kwargs[key])
    
    def try_get_attr(self, var_name, default_value):
        return getattr(self, var_name) if hasattr(self, var_name) else default_value
        
    def set_account_info(self, env:dict):
        account_info = env["account_info"] if "account_info" in env else {}
        if len(account_info)==0:
            self.username = ""
            self.auto_login = False
            self.auto_mfa_login = False
            self.mfa_secret = ""
            return
        self.username = account_info["username"]
        self.password = account_info["password"]
        self.mfa_secret = account_info["mfa_secret"] if "mfa_secret" in account_info else ""
        self.auto_login = (account_info["auto_login"]==1) if self.username!="" else False
        self.auto_mfa_login = (account_info["mfa_login"]==1) & self.auto_login if self.mfa_secret!="" else False
    
    def set_browser_settings(self, env:dict):
        self.new_browser = int(env['new_browser']) if 'new_browser' in env else 0
        self.attach_port = int(env['attach_port']) if 'attach_port' in env else 9222
        self.headless = bool(env['headless']) if 'headless' in env else False
        assert self.new_browser in [0,1,2], f"new_browser type {self.new_browser} not valid, check env variable."
        browser_name = env['browser'] if 'browser' in env else 'chrome'
        self.browser = browser_name.lower()
        self.new_browser = self.new_browser if self.browser=="chrome" else 1
        use_profile = False if 'useProfile' not in env else bool(env['useProfile'])
        profile_path = None if 'profilePath' not in env or len(env['profilePath'])==0 else env['profilePath']
        self.browser_info = {
            'name': browser_name,
            'use_profile': use_profile,
            'profile_path' : profile_path
        }

    def screenshot_to_output(self, filename, full_page:bool=False):
        dir_path = os.getcwd()
        dir_path = os.path.join(dir_path, self.task_output_dir)
        file_path = os.path.join(dir_path, filename)
        self.screenshot_to_dest(file_path, full_page)

    def save_json_file_to_output(self, filename, data:dict):
        save_json_file(data, self.task_output_dir, filename)
    
    def move_latest_download(self, download_dir:str, target_filename:str, limits=60, tolerant=False)->str:
        checkfile_ok, checkfile_wait, checkfile_fail = 1,2,0 
        def check_file(filename):
            file_ext = os.path.splitext(filename)[1]
            try:
                filets = os.path.getctime(filename)
                if time.time() - filets > limits:
                    self.logger.info(f"Latest file {filename} is too old.")
                    return checkfile_fail
            except FileNotFoundError as e:
                self.logger.info(f"Latest file {filename} is gone. {e}")
                return checkfile_wait
            if file_ext == '.crdownload' or file_ext == '.part':
                self.logger.info(f"Download has not finished...wait 30s and retry")
                return checkfile_wait
            return checkfile_ok
        def copy_ext(filename, target_filename):
            file_ext = os.path.splitext(filename)[1]
            return target_filename.rsplit('.', 1)[0] + file_ext
        download_dir = os.path.abspath(download_dir)
        if os.path.isdir(download_dir) == False or len(os.listdir(download_dir))==0:
            self.logger.info(f"no file in {download_dir}")
            return ""
        find_new_file = lambda x: max([os.path.join(x, f) for f in os.listdir(x)], key=os.path.getctime)
        filename = find_new_file(download_dir)
        retry, limit = 0, 10
        while (retry < limit) and (check_file(filename) == checkfile_wait):
            self.sleep_long()
            retry = retry + 1
            filename = find_new_file(download_dir)
        if check_file(filename) != checkfile_ok:
            self.logger.info(f"Download_dir: {download_dir}. File not found. ")
            if tolerant==False:
                raise FileDownloadError(f"File to download not found in target location.")
            else:
                self.logger.warning(f"Download file failed. target:{target_filename}")
                return ""
        else:
            self.logger.info(f"File download finished:{filename}")
        target_filename = copy_ext(filename, target_filename)
        target_filename = os.path.abspath(target_filename)
        create_dir_for_file(target_filename)
        shutil.move(filename, target_filename) 
        return target_filename

    def get_report_hash_key(self):
        return ["report_module", "report_class", "date_range", "site"]

    def get_report_hash(self, params:dict):
        def generate_report_name(task_param:dict, char_length=8):
            values = [str(value) for value in task_param.values()]
            str_all = "_".join(values)
            hash_object = hashlib.sha512(str_all.encode())
            hash_hex = hash_object.hexdigest()
            return hash_hex[0:char_length]
        key_columns =  self.get_report_hash_key()
        key_params = {key: value for key, value in params.items() if key in key_columns}
        return generate_report_name(key_params)

    def get_report_module(self, task_params:dict)->str:
        """
        Default to taskname.
        """
        return task_params['report_module'] if 'report_module' in task_params else self.taskname
    
    def get_report_class(self, task_params:dict)->str:
        """
        Default to taskname.
        """
        return task_params['report_class'] if 'report_class' in task_params else self.taskname

    def get_report_period(self, date_range:str, shift:int=0):
        start_date,_ = self.get_date_range_dates(date_range, shift)
        return start_date[:7]
    
    def get_date_range_dates(self, date_range:str, shift:int=0):
        """
        Get date range dates, return start_date, and end_date in "%Y-%m-%d" format
        """
        current = datetime.datetime.now()
        if self.execution_date:
            try:
                current = datetime.datetime.strptime(self.execution_date,"%Y-%m-%d")
            except ValueError:
                self.logger.error(f"execution_date: {self.execution_date} is not valid and not used.")
        return get_date_range_by_param(current, date_range, shift) 

    def pre_process_params(self, params:dict):
        if params is None:
            return {}
        if "date_range" in params:
            start_date, end_date =  self.get_date_range_dates(params["date_range"])  
            params['start_date'] = start_date
            params['end_date'] = end_date
            params["report_period"] = self.get_report_period(params["date_range"])
        params["report_module"] = self.get_report_module(params)
        params["report_class"] = self.get_report_class(params)
        params["report_name"] = self.get_report_hash(params)
        return params
    
    def go_to_site(self, site:str):
        self.logger.info(f"Go to site: {site}")
    
    def before_run_reports(self, report_params:dict):
        self.logger.info(f"Before run reports: {report_params.keys()}")
        return report_params

### implement task process steps -------------------------------
    def run_task_retry_handler(self, task_context:dict):
        restore_url = task_context["restore_url"] if "restore_url" in task_context else ""
        self.logger.info(f"Before retry, go back to start page. url: {restore_url}")
        if restore_url:           
            self.go_to(restore_url)
            
    def skip_retry_check(self, ex:Exception):
        critial_erros = ["WindowNotFound","NoOpenBrowser","BrowserNotFoundError","NoSuchWindowException",
        "NotLoggedInError","ProgramError","ParameterError"]
        skip_retry_errors = ["NoDownloadLinkError","ReportOptionError",
        "NoReportError","ParameterError","TooManyExceptionsError"]
        ex_name = ex.__class__.__name__ 
        if ex_name in critial_erros:
            return True
        if ex_name in skip_retry_errors:
            return True
        return False

## ----------------- error_handler decorator--------------------
    def error_handler(func):
        def decorate(self,*args, **kwargs):
            try:
                return func(self, *args, **kwargs)
            except Exception as e:
                self.handle_error(e)
                raise e
        return decorate

## ----------------- retry decorator--------------------
    def retry(func):
        def decorate(self,*args, **kwargs):
            c = 3
            while c > 0:
                try:
                    return func(self, *args, **kwargs)
                except Exception as e:
                    c -= 1
                    if self.skip_retry_check(e) == True:
                        c = 0
                    if c == 0:
                        raise e 
                    self.run_task_retry_handler(*args, **kwargs)  
        return decorate 

    @error_handler
    def before_run(self, task_context:dict)->dict:
        if hasattr(self, "account") == False:
            self.account = ""
        assert hasattr(self, "account"), "account is missing from env variables."
        assert hasattr(self, "home_url")
        account = self.account
        task_context["account"] = account
        self.logger.info(f"Start task with account: {account}")
        report_params = task_context["all_task_params"]
        task_context["all_task_params"] = self.before_run_reports(report_params)
        self.prepare_browser(self.home_url)
        return task_context

    @error_handler   
    def process_task_params(self, task_context:dict)->dict:
        report_params = task_context["all_task_params"]
        result = {}
        for site, site_reports in report_params.items():
            self.logger.info(f"process task params for site: {site}")
            site_combinations = []
            date_range_last = ""
            for report in site_reports:
                new_item = report.copy()
                new_item["site"] = site
                new_item = self.pre_process_params(new_item)
                site_combinations.append(new_item)
                date_range_last = new_item["date_range"] if "date_range" in new_item else date_range_last
            result[site] = site_combinations
        task_context["all_task_params"] = result
        task_context["report_period"] = self.get_report_period(date_range_last) if date_range_last else self.report_period
        return task_context
    
    @error_handler
    def switch_site(self, task_context:dict)->dict:
        site = task_context["site"]
        self.go_to_site(site)
        task_context["restore_url"] = self.location
        return task_context

    @error_handler
    def before_task_run(self, task_context:dict)->dict:
        return task_context

    @retry
    @error_handler
    def run_task(self, task_context:dict)->dict:
        taskmethod  = task_context["taskmethod"]
        params = task_context["task_params"]
        func = getattr(self, taskmethod)
        paths = func(params)
        task_result = params.copy()
        task_result["path"] = paths if isinstance(paths, list) else [paths]
        task_context["task_status"] = "success"
        task_context["task_result"] = task_result
        return task_context

    @error_handler
    def after_task_run(self, task_context:dict)->dict:
        return task_context

    @error_handler
    def task_end(self, task_context:dict)->dict:
        self.cleanup()
        return task_context
    
    @error_handler
    def after_run(self, task_context:dict)->dict:
        return task_context

### ------------------------------------------------------------
    def handle_error(self, ex):
        error_id = int(time.time())
        self.logger.info(f"Error track id: {error_id}")
        self.logger.error(ex, exc_info=True, stack_info=True)
        filename = f"error_{error_id}.jpg"
        try:
            self.screenshot_to_output(filename, full_page=False)
        except Exception as e:
            self.logger.error(f"Error taking screenshot, target filename:{filename}")
            self.logger.error(e, exc_info=True)
    



import datetime

def get_day_range(current:datetime.datetime, day_shift:int, duration:int):
    duration = duration if duration > 0 else 1
    today = current.replace(hour=0, minute=0, second=0, microsecond=0)
    start_date = today + datetime.timedelta(days=day_shift)
    end_date = start_date + datetime.timedelta(days=duration-1)
    return start_date, end_date

def datetime_to_str(dates):
    return tuple([dt.strftime("%Y-%m-%d") for dt in dates])

def normalize_date_range(date_range:str)->str:
    result = date_range.replace("-", "")
    result = result.replace("_", "")
    result = result.lower()
    return result

def parse_date_range(date_range):
    """
    LAST_MONTH@1
    """
    if "@" in date_range:
        date_range, shift = date_range.split("@")
        shift = int(shift)
    else:
        shift = 0
    return date_range, shift

def get_month_dates_by_param(date_range:str, shift=0):
    """
    Obsolete, to be deleted.
    """
    today = datetime.datetime.today()
    return get_date_range_by_param(today, date_range, shift)

def get_date_range_by_param(today:datetime.datetime, date_range:str, shift=0):
    date_range, shift_in_line = parse_date_range(date_range)
    shift = shift_in_line + shift
    date_range = normalize_date_range(date_range)
    dates_map = {
        "lastday": datetime_to_str(get_day_range(today,-1-shift,1)),
        "lastweek": get_week_range(today, -1-shift),
        "thismonth": get_this_month_range(today), 
        "lastmonth": get_month_valid_range(today, -1-shift),
        "lastyear": get_date_range_for_last_year(today),
        "last3months": get_date_range_for_last_n_months(today, 3),
        "last6months": get_date_range_for_last_n_months(today, 6,-shift)    
    }
    if date_range in dates_map:
        return dates_map[date_range]
    else:
        return get_date_range_for_days(today, date_range)

def get_date_range_for_last_year(current:datetime.datetime):
    today = current
    start_date = today.replace(year=today.year-1, day=1)
    end_date = today.replace(day=1) - datetime.timedelta(days=1)
    return start_date.strftime("%Y-%m-%d"), end_date.strftime("%Y-%m-%d")

def get_date_range_for_last_n_months(current:datetime.datetime, n:int, shift:int=0):
    today = current
    if shift>0:
        shift=0
    for i in range(-shift):
        today = last_month(today)
    start_date = today.replace(day=1)
    for i in range(n):
        start_date = last_month(start_date)
    end_date = today.replace(day=1) - datetime.timedelta(days=1)
    return start_date.strftime("%Y-%m-%d"), end_date.strftime("%Y-%m-%d")    

def get_date_range_for_days(current:datetime.datetime, days:str):
    if days.isdigit():
        days = int(days)
    else:
        days = 0
    # today = datetime.datetime.today()
    start_date = current - datetime.timedelta(days=days)
    return start_date.strftime("%Y-%m-%d"), current.strftime("%Y-%m-%d")

def next_month(dt):
    return (dt.replace(day=1) + datetime.timedelta(days=32)).replace(day=1)

def last_month(dt):
    return (dt.replace(day=1) - datetime.timedelta(days=2)).replace(day=1)

def get_report_period(date_range:str, execution_date:str)->str:
    current = datetime.datetime.now()
    if execution_date:
        try:
            current = datetime.datetime.strptime(execution_date,"%Y-%m-%d")
        except ValueError:
            pass
    start_date, _ = get_date_range_by_param(current, date_range)
    return start_date[:7]

def get_week_range_for_month(dt)->list:
    '''
    get week range for a month, from the last Monday to the next Sunday. 
    Make sure all weeks included for current month.
    '''
    day1 = dt.replace(day=1)
    next_month_day1 = next_month(dt)
    last_monday = day1 + datetime.timedelta(days=(7-day1.weekday())%7, weeks=-1)
    next_sunday = next_month_day1 + datetime.timedelta(days=-(next_month_day1.weekday()+1), weeks=1)
    weeks = []
    day = last_monday
    while day < next_sunday:
        weeks.append([day, day+datetime.timedelta(days=6)])
        day = day + datetime.timedelta(weeks=1)
    return weeks

def get_week_range(current:datetime.datetime, shift=0, week_start=-1):
    # current = datetime.datetime.today()
    if shift > 0:
        for i in range(1,shift+1):
            current = current + datetime.timedelta(weeks=1)
    elif shift < 0:
        for i in range(shift,0):
            current = current - datetime.timedelta(weeks=1)
    start_date = current - datetime.timedelta(days=current.weekday())+datetime.timedelta(days=week_start)
    end_date = start_date + datetime.timedelta(days=6)
    return start_date.strftime("%Y-%m-%d"), end_date.strftime("%Y-%m-%d")

def get_month_valid_range(current:datetime.datetime, shift=0):
    if shift > 0:
        shift = 0
    # current = datetime.datetime.today()
    for i in range(shift,0):
        current = last_month(current)
    start_date = current.replace(day=1).strftime("%Y-%m-%d")
    end_date_value = next_month(current) - datetime.timedelta(days=1)
    end_date_value = datetime.datetime.today() if end_date_value > datetime.datetime.today() else end_date_value
    end_date = end_date_value.strftime("%Y-%m-%d")
    return start_date, end_date

def get_this_month_range(current:datetime.datetime, shift:int=1):
    """
    Shift days for this month.
    """
    # current = datetime.datetime.today()
    return get_month_valid_range(current, -1) if current.day <= shift else get_month_valid_range(current)  

def get_month_dates(current:datetime.datetime, shift=0):
    current = current.replace(day=1)
    if shift >0:
        for i in range(1,shift+1):
            current = next_month(current)
    elif shift <0:
        for i in range(shift,0):
            current = last_month(current)
    start_date = current.strftime("%Y-%m-%d")
    end_date = (next_month(current) - datetime.timedelta(days=1)).strftime("%Y-%m-%d")
    return start_date, end_date

def get_date_format(locale:str):
    locale = locale.replace("-", "_")
    format_map = {
        "en_US": "%m/%d/%Y",
        "en_CA":"%Y-%m-%d",
        "en_DE": "%d/%m/%Y",
        "en_GB": "%d/%m/%Y"
    }
    if locale not in format_map:
        locale = "en_US" 
    return format_map[locale] 